********************************************************
* (C) Copyright 1987 - John Tonry. All Rights Reserved *
********************************************************
* Program to send a MONGO vector file to a PostScript laser printer
      parameter (nin=2048, nbuf=1048576, laserwidth=2550)
      parameter (maxpath=1024)
      integer*2 sign, buf(nbuf)
      character*80 line
      data sign /-32768/
      data iunit /6/

1000  format(a)
1001  format(2i5,1x,a)
1002  format(i4,1x,a)
1003  format(1x,a,i1)

******************************************************************************
* Vector file contains pairs of 16 bit numbers according to:                 *
*    Sign 1   Sign 2   Mantissa 1   Mantissa 2   Meaning                     *
*      0        0          x            y        Draw a dot at (x,y)         *
*      1        0          x            y        Relocate at (x,y)           *
*      0        1          x            y        Draw to (x,y)               *
*      1        1        32767        32767      End of vector file          *
*      1        1          0            0        Finish plot, start new page *
*      1        1          1            n        Line weight = n             *
*      1        1          2            n        Line type = n               *
* (x,y) are oriented in portrait mode                                        *
******************************************************************************

* Read in list of vectors and find how many there are
      read(5,1000) line
      open (unit=1,file=line,form='unformatted',status='old')
      ncoord = 0
      do 1 j = 1,nbuf/nin
          read(1,end=2) (buf(i+ncoord),i=1,nin)
          ncoord = ncoord + nin
1     continue
2     continue

* Initialize the laser printer
      write(iunit,1000) '%!PS-Adobe-1.0'
      write(iunit,1000) '%%EndComments'
      write(iunit,1000) 'save 50 dict begin /psplot exch def'
      write(iunit,1000) '/StartPSPlot'
      write(iunit,1000) '   {newpath 0 0 moveto 0 setlinewidth '//
     1	'0 setgray 1 setlinecap'
      write(iunit,1000) '    1 setlinejoin 72 300 div dup scale}def'
      write(iunit,1000) '/pending {false} def'
      write(iunit,1000) '/finish {pending {currentpoint stroke '//
     1	'moveto /pending false def} if} def'
      write(iunit,1000) '/r {finish newpath moveto} def'
      write(iunit,1000) '/d {lineto /pending true def} def'
      write(iunit,1000) '/l {finish 4 2 roll moveto lineto '//
     1	'currentpoint stroke moveto} def'
      write(iunit,1000) '/p {finish newpath moveto currentpoint '//
     1	'lineto currentpoint stroke moveto} def'
      write(iunit,1000) '/e {finish gsave showpage grestore newpath '//
     1	'0 0 moveto} def'
      write(iunit,1000) '/lw {finish setlinewidth} def'
      write(iunit,1000) '/lt0 {finish [] 0 setdash} def'
      write(iunit,1000) '/lt1 {finish [3 5] 0 setdash} def'
      write(iunit,1000) '/lt2 {finish [20 10] 0 setdash} def'
      write(iunit,1000) '/lt3 {finish [60 10] 0 setdash} def'
      write(iunit,1000) '/lt4 {finish [3 10 20 10] 0 setdash} def'
      write(iunit,1000) '/lt5 {finish [3 10 60 10] 0 setdash} def'
      write(iunit,1000) '/lt6 {finish [20 10 60 10] 0 setdash} def'
      write(iunit,1000) '/EndPSPlot {clear psplot end restore}def'
      write(iunit,1000) '% end fixed prolog'
      write(iunit,1000) 'StartPSPlot'
      write(iunit,1000) '%%EndProlog'
      write(iunit,1000) '%%Page: 1 1'

* Run through the vectors, sending them to the printer
* Note that several passes may be necessary for multipage plots
      maxx = 0
20    continue
      minx = maxx
      maxx = maxx + laserwidth
      n = 0
      more = 0
      ndraw = 0
      do 10 i = 1,ncoord,2
          i1 = and(sign,buf(i))
          m1 = buf(i) - i1
          i2 = and(sign,buf(i+1))
          m2 = buf(i+1) - i2
* Test to see whether the vector is on the present page
          if(i1.eq.0.or.i2.eq.0) then
              if(m1.gt.maxx) then
                  more = 1
                  goto 10
              else if(m1.lt.minx) then
                  goto 10
              end if
          end if
* Write the output line if it will exceed the length allotted
          if(n.gt.(80-12)) then
              write(iunit,1000) line(:n)
              n = 0
          end if
* point or draw?
          if(i1.eq.0) then
* point at m1, m2
              if(i2.eq.0) then
                  write(line(n+1:n+12),1001) m1-minx, m2, 'p'
                  n = n + 12
                  ndraw = 0
              else
* draw to m1, m2
                  write(line(n+1:n+12),1001) m1-minx, m2, 'd'
                  n = n + 12
                  ndraw = ndraw + 1
* Check to see if our path is too long
                  if(ndraw.ge.maxpath) then
                      ndraw = 0
* Write the output line if it will exceed the length allotted
                      if(n.gt.(80-12)) then
                          write(iunit,1000) line(:n)
                          n = 0
                      end if
                      write(line(n+1:n+12),1001) m1-minx, m2, 'r'
                      n = n + 12
                  end if
              end if
* relocate or special?
          else
* relocate at m1, m2
              if(i2.eq.0) then
                  write(line(n+1:n+12),1001) m1-minx, m2, 'r'
                  n = n + 12
                  ndraw = 0
* end of buffer
              else if(m1.eq.32767) then
                  goto 11
* eject page, start new one
              else if(m1.eq.0) then
                  if(n.gt.0) write(iunit,1000) line(:n)
                  write(iunit,1000) 'e'
                  n = 0
                  ndraw = 0
* set line weight
              else if(m1.eq.1) then
                  write(line(n+1:n+7),1002) m2, 'lw'
                  n = n + 7
                  ndraw = 0
* set line type
              else if(m1.eq.2.and.m2.ge.0.and.m2.le.6) then
                  write(line(n+1:n+4),1003) 'lt',m2
                  n = n + 4
                  ndraw = 0
              end if
          end if
10    continue

11    continue
      if(more.eq.0) goto 999
      if(n.gt.0) write(iunit,1000) line(:n)
      write(iunit,1000) 'e'
      goto 20

999   continue
* Write the LINE buffer if it is not empty and print the page
      if(n.gt.0) write(iunit,1000) line(:n)
      write(iunit,1000) 'e'

* Finish up the plot
      write(iunit,1000) '%%Trailer'
      write(iunit,1000) 'EndPSPlot'

      stop
      end

      function nch(s)
      character*(*) s
      do 900 i = len(s),1,-1
          nch = i
          if(s(i:i).ne.' '.and.ichar(s(i:i)).ne.0) return
900   continue
      return
      end
