/*
 * Initialize X server connection and create a base window
 * JT Sept 1998
 */
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#define ERR_CANT_OPEN_DISPLAY			1
#define ERR_CANT_CREATE_IMAGE_WINDOW		2
#define ERR_INSUFF_IMAGE_COLORS			6

/*#define TEST /* */

#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define ABS(a) (((a) > 0) ? (a) : -(a))

#define MAXCLASS 6
struct mgclasstruct {
  int class;
  char *name;
} mgclassname[MAXCLASS] = {
  StaticGray,   "StaticGray", 
  GrayScale,  	"GrayScale",  
  StaticColor,	"StaticColor",
  PseudoColor,	"PseudoColor",
  TrueColor,  	"TrueColor",  
  DirectColor,	"DirectColor"};

mgxopen(display, progname, dp, scrn, cmap, owncmap, nclr, pixels, 
      back, fore, visual, vclass, vdepth, w, h, win, xswa)
char *display;		/* name of X server */
char *progname;		/* name of program (for resource lookup) */
Display **dp;		/* display of X server */
int *scrn;		/* screen of X server (-1 => default) */
Visual **visual;	/* Visual used in case of multiples */
int  *vclass;		/* Visual class (-1 => default) */
int  *vdepth;		/* Visual depth */
Colormap *cmap;		/* color map used */
int *owncmap;		/* 0/1 for own color map */
int *nclr;		/* requested and returned number of color cells */
unsigned long *pixels;	/* color cell addresses allocated */
unsigned long *back;	/* background pixel */
unsigned long *fore;	/* foreground pixel */
int *w;			/* requested and returned width of window */
int *h;			/* requested and returned height of window */
Window *win;		/* Base window */
XSetWindowAttributes *xswa;	/* Window attributes */
{
  int i, nbad;
  int rmask, gmask, bmask;
  unsigned long int planes;
  char *screenstring, *geomstring, *backstring, *forestring, *visualstring;
  Colormap tmpcmap;

  int *listdepths, nvis, defclass;
  XVisualInfo *visinfo, vinfo_template;

  XSizeHints size_hints;
  XWMHints wm_hints;
  int result;
  int border_width=2;
  XTextProperty	iconname, winname;
  char	*strtmp,stmp[80];
  Window inforoot;
  int infox, infoy, infowidth, infoheight, infoborder, infodepth;
  XColor backcolor;
  unsigned long linuxbug[3];	/* color cell addresses wrongly allocated */

#ifdef TEST
  fprintf(stderr,"OK, time to open connection\n");
#endif

/* Open the connection to the server, quit on failure */
  if( ! (*dp = XOpenDisplay(display)) ) {
    fprintf(stderr,"Can't open display %s\n",XDisplayName(display));
    return(ERR_CANT_OPEN_DISPLAY);
  }

#ifdef TEST
  XSynchronize(*dp, True);
#endif

/* Select a screen number from X, then from default, then from request */
  if(*scrn < 0) {
    screenstring = XGetDefault(*dp, progname, "screen");
    if(screenstring != NULL) *scrn = atoi(screenstring);
    else *scrn = DefaultScreen(*dp);
  }
  
#ifdef TEST
  fprintf(stderr,"Selected display %x and screen %d\n", *dp, *scrn);
#endif

/* If default visual, see if there is an Xdefault, otherwise use default */
  if(*vclass < 0) {
    visualstring = XGetDefault(*dp, progname, "visual");
    if(visualstring != NULL) {
      for(i=0; i<MAXCLASS; i++) {
	if(strcasecmp(visualstring, mgclassname[i].name) == 0)
	  *vclass = mgclassname[i].class;
      }
    }
  }
  
/* Look for a visual of the appropriate class */
  visinfo = XGetVisualInfo(*dp, 0L, &vinfo_template, &nvis);

/* Does this screen even exist? */
  *visual = NULL;
  for(i=0; i<nvis; i++) {
    if((visinfo+i)->screen == *scrn) {
      *visual = (visinfo+i)->visual;
      break;
    }
  }
  if(*visual == NULL) {
    printf("Screen %d doesn't exist,", *scrn);
    *scrn = DefaultScreen(*dp);
    printf(" using default screen %d.\n", *scrn);
  }

  *visual = NULL;
  if(*vclass >= 0) {
    if(visinfo != NULL) {
#ifdef TEST
      fprintf(stderr, "Visual Info: (Gray = %d, Pseudo = %d, True = %d, Direct = %d)\n",
	      GrayScale, PseudoColor, TrueColor, DirectColor);
      fprintf(stderr,
	      "    Visual   Visid scrn dpth class r_mask g_mask b_mask size bits\n");
#endif
      for(i=0; i<nvis; i++) {
	if(*visual == NULL &&
	   (visinfo+i)->class == *vclass && (visinfo+i)->screen == *scrn &&
/* 8-bit true color isn't worth having */
	   ((visinfo+i)->class != TrueColor || (visinfo+i)->depth > 8)) {
	  *visual = (visinfo+i)->visual;
	  *vclass = (visinfo+i)->class;
	  *vdepth = (visinfo+i)->depth;
	  rmask = (visinfo+i)->red_mask;
	  gmask = (visinfo+i)->green_mask;
	  bmask = (visinfo+i)->blue_mask;
	}
#ifdef TEST
	fprintf(stderr, "%10x %8d %4d %4d %4d %6x %6x %6x %4d %4d\n",
		(visinfo+i)->visual,
		(visinfo+i)->visualid,
		(visinfo+i)->screen,
		(visinfo+i)->depth,
		(visinfo+i)->class,
		(visinfo+i)->red_mask,
		(visinfo+i)->green_mask,
		(visinfo+i)->blue_mask,
		(visinfo+i)->colormap_size,
		(visinfo+i)->bits_per_rgb);
#endif
      }
      if(*visual == NULL) {
	printf("Can't find %s visual,", mgclassname[*vclass].name);
	*visual = DefaultVisual(*dp, *scrn);
	for(i=0; i<nvis; i++) {
	  if((visinfo+i)->visual == *visual) {
	    *vclass = (visinfo+i)->class;
	    *vdepth = (visinfo+i)->depth;
	    rmask = (visinfo+i)->red_mask;
	    gmask = (visinfo+i)->green_mask;
	    bmask = (visinfo+i)->blue_mask;
	  }
	}
	printf(" using default %s.\n", mgclassname[*vclass].name);
      }
    } else {
      fprintf(stderr, "Can't get visual info!\n");
    }
  } else {
    *visual = DefaultVisual(*dp, *scrn);
    for(i=0; i<nvis; i++) {
      if((visinfo+i)->visual == *visual) {
	*vclass = (visinfo+i)->class;
	*vdepth = (visinfo+i)->depth;
	rmask = (visinfo+i)->red_mask;
	gmask = (visinfo+i)->green_mask;
	bmask = (visinfo+i)->blue_mask;
      }
    }
  }
  XFree(visinfo);

#ifdef TEST
  fprintf(stderr, "Default visual  %10x\n", DefaultVisual(*dp, *scrn));
  fprintf(stderr, "Selected visual %10x, Class %d, Depth %d\n", *visual,*vclass,*vdepth);
  if(*visual != DefaultVisual(*dp, *scrn)) {
    fprintf(stderr,"Not the default visual, might well fail...\n");
  }
#endif

/* Get a color map */
  if(*visual == DefaultVisual(*dp, *scrn) && *owncmap == 0) {
#ifdef TEST
    fprintf(stderr,"Trying for default color map\n");
#endif
    if ((*cmap = DefaultColormap(*dp, *scrn))==0) {
      fprintf(stderr,"Can't get color map from screen %d of display %s\n",
	      *scrn, XDisplayName(display));
      return ERR_CANT_OPEN_DISPLAY;
    }
#ifdef TEST
    fprintf(stderr,"Got default color map\n");
#endif

/* Make sure we can get nclr colors from the server */
    if(*vclass == PseudoColor) {
#ifdef TEST
      fprintf(stderr,"Trying for %d colors\n", *nclr);
#endif
      if (XAllocColorCells(*dp, *cmap,
			   0, &planes, 0, pixels, *nclr) == 0) {
/* Here is the place to try again with fewer colors requested */
	fprintf(stderr,"Insufficient colors for cmap, try own map\n");
	*owncmap = 1;
      }
    }
  } else if(*vclass == PseudoColor) {
    *owncmap = 1;
  }

/* Stash these away in pixels[] since it is unused for TrueColor */
  if(*vclass == TrueColor) {
    pixels[0] = rmask;
    pixels[1] = gmask;
    pixels[2] = bmask;
    if ((*cmap = XCreateColormap(*dp, RootWindow(*dp, *scrn),
				   *visual, AllocNone)) == 0) {
      fprintf(stderr,"Can't get colormap from screen %d of display %s\n", *scrn,
	      XDisplayName(display));
      return ERR_CANT_OPEN_DISPLAY;
    }
  }

  if(*owncmap && *vclass == PseudoColor) {
    fprintf(stderr,"Trying alternate PseudoColor visual from screen %d of display %s\n",*scrn, XDisplayName(display));
    if ((*cmap = XCreateColormap(*dp, RootWindow(*dp, *scrn),
				   *visual, AllocNone)) == 0) {
      fprintf(stderr,"Can't get own map from screen %d of display %s\n", *scrn,
	      XDisplayName(display));
      return ERR_CANT_OPEN_DISPLAY;
    }
/* Ask for the nclr colors from the server */
    if(*vclass == PseudoColor) {
      if (XAllocColorCells(*dp, *cmap,
			   0, &planes, 0, pixels, *nclr) == 0) {
/* Here is the place to try again with fewer colors requested, */
/* maybe falling back to BlackPixel and WhitePixel */
	fprintf(stderr,"Insufficient colors for cmap\n");
	return(ERR_INSUFF_IMAGE_COLORS);
      }
/*
 * This is because the RedHat Xaccel server seems to allocate pixel 2 in a 
 * private colormap, but never actually relinquish use of it.  The result
 * is a color cell which can apparently be set and read, but which always
 * displays as gray (color of border?).
 * Therefore check that cells 0-2 are not among those returned!
 */
      for(i=0, nbad=0; i<*nclr; i++) {
	if(pixels[i] <= 2) {
	  linuxbug[nbad++] = pixels[i];
	  if (XAllocColorCells(*dp, *cmap,
			   0, &planes, 0, &pixels[i], 1) == 0) {
	    fprintf(stderr,"Insufficient colors for cmap\n");
	    return(ERR_INSUFF_IMAGE_COLORS);
	  }
	}
      }
      if(nbad > 0) {
	fprintf(stderr,"X server bug?  Freeing %d colors which probably shouldn't have been allocated.\n", nbad);
	XFreeColors(*dp, *cmap, linuxbug, nbad, 0);
      }
    }
  }


/* See if we have a request for a background color */
  backcolor.flags = DoRed | DoGreen | DoBlue;
/* Nope, these need to be writeable cells */
/*
  if( (backstring = XGetDefault(*dp, progname, "background")) == NULL &&
      *cmap == DefaultColormap(*dp, *scrn) ) {
    *back = BlackPixel(*dp, *scrn);

  } else {
*/
  if( (backstring = XGetDefault(*dp, progname, "background")) == NULL || 
      XParseColor(*dp, *cmap, backstring, &backcolor) == 0)
    backcolor.red = backcolor.green = backcolor.blue = 0;     /* black */
  
  if(*vclass == PseudoColor) {
    if(XAllocColorCells(*dp, *cmap, 0, &planes, 0, &backcolor.pixel, 1) == 0)
      fprintf(stderr,"Insufficient colors for background...\n");
  }
  mgxsetcolor(*dp, *vclass, *cmap, &backcolor);
  *back = backcolor.pixel;

#ifdef TEST
  fprintf(stderr,"bckg = %5d %8d %8d %8d\n", backcolor.pixel, 
	  backcolor.red, backcolor.green, backcolor.blue);
#endif
/*  }  */

/* See if we have a request for a foreground color */
/* Nope, these need to be writeable cells */
/*
  if( (forestring = XGetDefault(*dp, progname, "foreground")) == NULL &&
      *cmap == DefaultColormap(*dp, *scrn) ) {
    *fore = WhitePixel(*dp, *scrn);

  } else {
*/
  if( (forestring = XGetDefault(*dp, progname, "foreground")) == NULL || 
      XParseColor(*dp, *cmap, forestring, &backcolor) == 0)
    backcolor.red = backcolor.green = backcolor.blue = 65535;  /* white */

  if(*vclass == PseudoColor) {
    if(XAllocColorCells(*dp, *cmap, 0, &planes, 0, &backcolor.pixel, 1) == 0)
      fprintf(stderr,"Insufficient colors for foreground...\n");
  }

  mgxsetcolor(*dp, *vclass, *cmap, &backcolor);
  *fore = backcolor.pixel;

#ifdef TEST
  fprintf(stderr,"forg = %5d %8d %8d %8d\n", backcolor.pixel, 
	  backcolor.red, backcolor.green, backcolor.blue);
#endif
/*  } */

#ifdef TEST
  fprintf(stderr,"Got %d colors\n", *nclr);
  fprintf(stderr, "Color cells allocated, cmap = %x (def = %x)\n", *cmap,
	  DefaultColormap(*dp, *scrn));

  for(i=0; i<*nclr; i++) 
    fprintf(stderr,"%3d%c", pixels[i], (i+1)%16==0?'\n':' ');
  fprintf(stderr,"\n");
#endif

/* Set the window attributes */
/*  xswa->override_redirect = True; */
/*  xswa->backing_store = WhenMapped;*/
  xswa->background_pixel = *back;
  xswa->border_pixel = *fore;
  xswa->colormap = *cmap;

/* Set window manager properties */
/* Set the window geometry if specified */
  size_hints.x = size_hints.y = 0;
  size_hints.width = *w;
  size_hints.height = *h;
  size_hints.min_width = size_hints.min_height = 256;
  size_hints.base_width = size_hints.base_height = 256;
  size_hints.max_width = size_hints.max_height = 512;
  size_hints.flags = PPosition | PSize | PMinSize | PBaseSize;
  geomstring = XGetDefault(*dp, progname, "geometry");
  if (geomstring!=NULL) {
    result = XParseGeometry(geomstring,&size_hints.x,&size_hints.y,
                            &size_hints.width,&size_hints.height);
    if (result & XNegative)
      size_hints.x += DisplayWidth(*dp, *scrn) - size_hints.width - 2*border_width;
    if (result & YNegative)
      size_hints.y += DisplayHeight(*dp, *scrn) - size_hints.height - 2*border_width;;
    if (result & XValue || result & YValue) {
      size_hints.flags &= ~PPosition;
      size_hints.flags |= USPosition;
    }
    if (result & WidthValue || result & HeightValue) {
      size_hints.flags &= ~PSize;
      size_hints.flags |= USSize;
    }
/* width and height set the upper bounds for auto image size */
    if (result & WidthValue) {
      *w = MIN(*w,size_hints.width);
    }
    if (result & HeightValue) {
      *h = MIN(*h,size_hints.height);
    }
  }

#ifdef TEST
  fprintf(stderr, "args = %d %d %d %d %d %d %d %d %d %d %d %d\n",
	  *dp, RootWindow(*dp, *scrn),
                        0,0, *w, *h, border_width,
                        *vdepth, InputOutput, *visual,
                        CWBackPixel | CWColormap, xswa);
#endif

  *win = XCreateWindow(*dp, RootWindow(*dp, *scrn),
                        size_hints.x, size_hints.y, *w, *h, border_width,
                        *vdepth, InputOutput, *visual,
                        CWBackPixel | CWBorderPixel | CWColormap, xswa);

  if (!*win) { 
    fprintf(stderr, "XCreateWindow failed\n");
    return(ERR_CANT_CREATE_IMAGE_WINDOW);
  }

#ifdef TEST
  fprintf(stderr, "Basewindow created...\n");
#endif

  XSetIconName(*dp, *win, progname);
  XStoreName(*dp, *win, progname);
  XSetWMNormalHints(*dp, *win, &size_hints);
#ifdef TEST
  fprintf(stderr, "WM props set...\n");
#endif

/* Map the window and see what size it is (in case user has resized it) */
  XMapWindow(*dp, *win);

#ifdef TEST
  fprintf(stderr, "Base window mapped, about to get geometry...\n");
#endif

/* Set width and height to the size of the window */
  XGetGeometry(*dp, *win, &inforoot,&infox,&infoy,&infowidth,&infoheight,
               &infoborder,&infodepth);
  *w = infowidth;
  *h = infoheight;

  return(0);
}

mgxsetcolor(dp, vclass, map, color)
Display *dp;
Colormap map;
XColor *color;
int vclass;
{
  switch(vclass) {
  case PseudoColor:
  case GrayScale:
  case DirectColor:
    XStoreColor(dp, map, color);
    break;
  case StaticGray:
  case StaticColor:
  case TrueColor:
    XAllocColor(dp, map, color);
    break;
  default:
    break;
  }
}
