CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
CC                  Mongo Interactive Graphics Software                    CC
CC                 Copyright (c) 1987, 1994 - John Tonry.                  CC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine mgohalftone(data,nx,ny,z1,z2,contrast,ialg)
* Program to make a gray-scale picture of a file
* maxbuf should be larger than # of horiz or vertical dots of all 
* output devices. For a 300 dpi laserprinter, 3150 should suffice
      parameter (maxbuf=3150)
      parameter (maxdens=4096, ngrid=16, npixsize=4)
      include 'mongo.par'
* Data are in DATA(NX,NY)
* There are MX x MY hardcopy pixels in the X x Y direction 
* The data is mapped to (0,1) using the array DENS(NDENS) and the linear
* map with f(DMIN) = 0 and f(DMAX) = 1
      real data(nx,ny)
      integer*2 pic(maxbuf), scratch(maxbuf+2)
      integer count(maxdens)
      integer*2 dens(maxdens)
      real fdens(maxdens)
      logical neg, linear
      character*80 line

      if(termout .or. ialg.eq.0) then
         ngray = 2
         ndens = maxdens
         mx = nint(gx2-gx1)
         my = nint(gy2-gy1)
      else
         ngray = 10
         ndens = (ngray-1)*(min(maxdens,256*(ngray-1))/(ngray-1))
         mx = nint((gx2-gx1)/npixsize)
         my = nint((gy2-gy1)/npixsize)
      end if

      if(ialg.eq.0) then
         nxstep = mx
         nystep = 1
* SCRATCH is the array of previous pixel remainders. Initialize it with random
* numbers between 0 and 1 in order to prevent obnoxious patterns in smooth data
         y = .5
         do 900 i = 1,mx+2
            scratch(i) = ndens*y
            y = pi*y - int(pi*y)
 900     continue
      else
         nxstep = ngrid
         nystep = ngrid
      end if

      if(mx.gt.maxbuf) then
          call mgoerr('HALFTONE: buffer space exceeded')
          return
      end if

      if(z1.eq.z2) then
          write(line,'(1p2g12.4)') z1, z2
          call mgoerr2('HALFTONE: zero data range...', line)
          return
      end if

      neg = z2.lt.z1

      if(neg) then
          dmin = z2
          dmax = z1
      else
          dmin = z1
          dmax = z2
      end if
      linear = abs(contrast) .gt. 99.9

* Compute the density table: output density for input index
      if(linear) then
          do 902 i = 1,ndens
              fdens(i) = float(i-1) / (ndens-1)
902       continue
* Derive an enhanced, non-linear density transformation by calling DENSITY
* In order to use the DENSITY subroutine count the data with COUNT
      else
          call mgocountup(nx, ny, data, dmin, dmax, ndens, count)
          call mgodensity(ndens, count, fdens, contrast)
      end if

* Print a table of the density values
c      call mgoprdens(dmin,dmax,ndens,dens,count)

* If a printer, adjust the density transformation to account for the 
* (overlarge) dot characteristics.
      if(.not.termout) then
         adjust = 2.5
         if(ialg.eq.0) adjust = 3.5
         call mgofatpixmap(ndens, fdens, adjust, neg)
      end if

* Create the lookup table:
      do 903 i = 1,ndens
          dens(i) = nint(ndens*fdens(i))
903   continue      

      call newmgobitinit(ialg, .true.)


      do 400 j0 = 0,my-1,nystep
         do 401 i0 = 0,mx-1,nxstep
* Map the current data line to density values using the density transformation
            call mgodensmap(nx,ny,data, dmin,dmax,ndens,dens,
     $           mx,my,i0,j0, nxstep,nystep,pic,isum)

* Use the line of densities to put out patches of half-tone dots
            if(ialg.eq.0) then
               call mgolinodots(mx,ndens,pic,scratch,neg)
            else
               call mgodiscrete(ndens,ngray,ngrid,pic,isum,neg)
               if(ngray.eq.2) call mgorelax(ngray,ngrid,pic)
            end if

            if(termout.or.ialg.eq.0) then
               call mgodrawbits(gx1+i0,gy1+j0,nxstep,nystep,pic)
            else
               call mgodrawbits(gx1+npixsize*i0,
     $              gy1+npixsize*j0,nxstep,nystep,pic)
            end if

 401     continue
 400  continue

      call newmgobitinit(ialg, .false.)

      return
      end

      subroutine mgodensmap(nx,ny,data, dmin,dmax,ndens,dens,
     $     mx,my, i0,j0, ni,nj,pic,isum)
* Converts real data DATA to density values from 0 to ndens, returning the
* results in PIC. The map is according to DMIN, DMAX, DENS(NDENS).
* The data values are in an array of dimension (nx,ny); we want to sample
* on a grid of dimension (mx,my)
* Fill a (ni,nj) array PIC, starting at (i0,j0) in (mx,my)
      real data(nx,ny)
      integer*2 dens(ndens), pic(ni,nj)
      adens = ndens / (dmax-dmin)
      isum = 0
      do 10 jm = 1,nj
         j = max(1,min(ny,nint((ny*float(jm-1+j0))/(my-1))))
         do 11 im = 1,ni
            i = max(1,min(nx,nint((nx*float(im-1+i0))/(mx-1))))
            t = data(i,j)
            idx = min(ndens,max(1,nint(adens*(t-dmin))))
            pic(im,jm) = dens(idx)
            isum = isum + pic(im,jm)
 11      continue
 10   continue
      return
      end

      subroutine mgolinodots(npix,ndens,data,buf,neg)
* Program converts a line of density values (0-ndens) in DATA to a bit for bit
* halftone picture in DATA. BUF is scratch space, and NEG is a flag to
* make a negative picture (unset highlights).
      integer*2 data(npix), buf(npix+2)
      logical ovfl, neg, increase
      data increase /.true./

      if(increase) then
          ndot = 1
          ndelta = 1
      else
          ndot = npix
          ndelta = -1
      end if
      iprev = 0

      do 901 i = 1,npix
          ivalue = data(ndot) +
     1       (buf(ndot) + iprev + buf(ndot+1) + buf(ndot+2))/4
          ovfl = ivalue .gt. ndens
          data(ndot) = 0
          if((.not.neg).eqv.ovfl) data(ndot) = 1
          if(ovfl) ivalue = ivalue - ndens
          iprev = buf(ndot+1)
          buf(ndot+1) = ivalue
          ndot = ndot + ndelta
901   continue
      increase = .not. increase

      return
      end

      subroutine mgodiscrete(ndens,ngray,ngrid,pic,isum,neg)
      integer*2 pic(ngrid,ngrid)
      integer ir(3,4)
      logical neg

      nunit = ndens/(ngray-1)
      icorrect = mod(isum,nunit)
      if(icorrect.gt.nunit/2) icorrect = icorrect - ndens

      call mgoadjust(-icorrect,ngrid,ngrid,pic)

      n = ngrid
      do 10 k = 1,100
         if(n.eq.1) goto 99
         do 11 j = 1,ngrid,n
            do 12 i = 1,ngrid,n
               call mgoresiduals(nunit,n,ngrid,pic,i,j,ir)
               do 13 m = 1,4
                  call mgoadjust(ir(3,m),ngrid,n/2,pic(ir(1,m),ir(2,m)))
 13            continue
 12         continue
 11      continue
         n = n / 2
 10   continue
 99   continue

      do 20 j = 1,ngrid
         do 21 i = 1,ngrid
            pic(i,j) = min(ngray-1,max(0,pic(i,j)/nunit))
            if(neg) pic(i,j) = ngray-1 - pic(i,j)
 21      continue
 20   continue
      
      return
      end

      subroutine mgoadjust(idelt,nx,n,pic)
      parameter (pi=3.1415926535)
      integer*2 pic(nx,1)
      data randum /0.5/
      z = (pi+randum)*(pi+randum)*(pi+randum)*(pi+randum)*(pi+randum)
      randum = z - int(z)
      if(idelt.gt.0) then
         delt = idelt + (0.2+0.6*randum)
      else
         delt = idelt - (0.2+0.6*randum)
      end if
      delt = delt / (n*n)
      add = 0
C      write(6,*) idelt, nx, n, pic(1,1), delt
      do 10 j = 1,n
         do 11 i = 1,n
            add = add + delt
            pic(i,j) = pic(i,j) + int(add)
            add = add - int(add)
 11      continue
 10   continue
      return
      end

      subroutine mgoresiduals(nunit,n,nx,pic,i0,j0,ir)
      integer*2 pic(nx,nx)
      integer irk(17)
      integer ir(3,4)
C      if(nsrt.eq.0) write(6,*) 'Hi there, rev 7'
      ir(1,1) = i0
      ir(2,1) = j0
      ir(3,1) = 0
      ir(1,2) = i0 + n/2
      ir(2,2) = j0
      ir(3,2) = 0
      ir(1,3) = i0
      ir(2,3) = j0 + n/2
      ir(3,3) = 0
      ir(1,4) = i0 + n/2
      ir(2,4) = j0 + n/2
      ir(3,4) = 0
      do 10 j = 0,n/2-1
         do 11 i = 0,n/2-1
            ir(3,1) = ir(3,1) + pic(ir(1,1)+i,ir(2,1)+j)
            ir(3,2) = ir(3,2) + pic(ir(1,2)+i,ir(2,2)+j)
            ir(3,3) = ir(3,3) + pic(ir(1,3)+i,ir(2,3)+j)
            ir(3,4) = ir(3,4) + pic(ir(1,4)+i,ir(2,4)+j)
 11      continue
 10   continue
      ir(3,1) = -mod(ir(3,1),nunit)
      ir(3,2) = -mod(ir(3,2),nunit)
      ir(3,3) = -mod(ir(3,3),nunit)
      ir(3,4) = -mod(ir(3,4),nunit)
      netres = -(ir(3,1) + ir(3,2) + ir(3,3) + ir(3,4))
      do 910 k = 2,4
         i = ir(1,k)
         j = ir(2,k)
         isum = ir(3,k)
         l = k - 1
 911     if(l.ge.1.and.ir(3,l).gt.isum) then
            ir(1,l+1) = ir(1,l)
            ir(2,l+1) = ir(2,l)
            ir(3,l+1) = ir(3,l)
            l = l - 1
            nsrt = nsrt + 1
            goto 911
         end if
         ir(1,l+1) = i
         ir(2,l+1) = j
         ir(3,l+1) = isum
 910  continue
      if(netres.ge.3*nunit) then
         ir(3,1) = ir(3,1) + nunit
         ir(3,2) = ir(3,2) + nunit
         ir(3,3) = ir(3,3) + nunit
      else if(netres.ge.2*nunit) then
         ir(3,1) = ir(3,1) + nunit
         ir(3,2) = ir(3,2) + nunit
      else if(netres.ge.nunit) then
         ir(3,1) = ir(3,1) + nunit
      end if

      return
      end

      subroutine mgorelax(ngray,n,pic)
      integer*2 pic(n,n)
      do 10 j = 1,n
         do 11 i = 1,n
            p0 = mgopotential(ngray,n,pic,i,j)
            ph = 1000
            pv = 1000
            isave = pic(i,j)
            if(i.gt.1) then
               pic(i,j) = pic(i-1,j)
               pic(i-1,j) = isave
               ph = mgopotential(ngray,n,pic,i-1,j)
               pic(i-1,j) = pic(i,j)
               pic(i,j) = isave
            end if
            if(j.gt.1) then
               pic(i,j) = pic(i,j-1)
               pic(i,j-1) = isave
               pv = mgopotential(ngray,n,pic,i,j-1)
               pic(i,j-1) = pic(i,j)
               pic(i,j) = isave
            end if
            if(p0.le.amin1(ph,pv)) goto 11
            if(ph.le.pv) then
               pic(i,j) = pic(i-1,j)
               pic(i-1,j) = isave
            else
               pic(i,j) = pic(i,j-1)
               pic(i,j-1) = isave
            end if

 11      continue
 10   continue
      return
      end

      function mgopotential(ng,n,pic,i,j)
      integer*2 pic(n,n)
      ip = i + 1
      if(ip.gt.n) ip = ip - n
      im = i - 1
      if(im.lt.1) im = im + n
      jp = j + 1
      if(jp.gt.n) jp = jp - n
      jm = j - 1
      if(jm.lt.1) jm = jm + n
      mgopotential = 
     $     (pic(i,j)-0.5*ng)*(pic(ip,j)-0.5*ng) + 
     $     (pic(i,j)-0.5*ng)*(pic(i,jp)-0.5*ng) + 
     $     (pic(i,j)-0.5*ng)*(pic(im,j)-0.5*ng) + 
     $     (pic(i,j)-0.5*ng)*(pic(i,jm)-0.5*ng)
      return
      end

      subroutine mgocountup(nx,ny,data,dmin,dmax,ndens,count)
* Counts the frequency of different values found in the array DATA. The
* actual data values are first mapped from 1 to NDENS according to a
* linear map that has f(DMIN) = 1 and f(DMAX) = NDENS, and this is then
* used as an index in the array COUNT.
      integer count(ndens)
      real data(nx,ny)
      adens = ndens / (dmax-dmin)
      do 10 i = 1,ndens
      count(i) = 0
10    continue
      do 20 j = 1,ny
      do 20 i = 1,nx
      t = adens*(data(i,j) - dmin)
      index = nint(t)
      index = min(max(1,index),ndens)
      count(index) = count(index) + 1
20    continue
      return
      end

      subroutine mgodensity(ndens,count,dens,darkness)
* Density computes a density transformation, using a little image
* enhancement (equal number of pixels at each density level).
      integer count(ndens)
      real dens(ndens)

C darkness < 0 is the (-) density at the (mode + fwhm) of the background peak
C darkness > 0 is the squashing factor of the density transformation
C at the (mode + fwhm) of the background peak

* Find the largest index in COUNT with non-zero count.
      do 30 maxbin = ndens,1,-1
      if(count(maxbin).ne.0) goto 31
30    continue
31    continue

* Add up the total number of pixels (for normalization) and also
* find the mode of the frequency distribution.
      sum = count(1)
      mode = 0
      maxcount = 0
* Loop from 2 in order to avoid spurious peaks at zero intensity
      do 40 i = 2,maxbin
      if(count(i).gt.maxcount) then
          maxcount = count(i)
          mode = i
      endif
40    sum = sum + count(i)

* Now find the width of the peak in the frequency distribution by running
* down from the mode until half-maximum
      ihwhm1 = 0
      ihwhm2 = 0
      do 45 i = 1,mode
      if(count(mode+i+1).lt.maxcount/2 .and.
     1    count(mode+i).gt.maxcount/2) then
          ihwhm2 = i
      endif
      if(count(mode-i+1).lt.maxcount/2 .and.
     1    count(mode-i+2).gt.maxcount/2) then
          ihwhm1 = i
      endif
      if(ihwhm1.gt.0.and.ihwhm2.gt.0) goto 46
45    continue
46    ifwhm = ihwhm1 + ihwhm2

* Compute the initial density transformation as the integral of the
* frequency distribution (to give equal number of pixels at each density)
      dens(1) = count(1)/sum
      do 50 j = 2,ndens
50    dens(j) = dens(j-1) + count(j)/sum

c	write(6,1001) mode, maxcount, ifwhm
c1001	format(1x,'Mode is at',I5,' with',I7,' pixels, fwhm =',I5)

* In a picture with a dark background (for example astronomical pictures)
* the dark pixels are very much over-represented, and do not contain any
* information. Thus, alter the density distribution by continuously
* squashing it down towards 0 or up towards 1.
* If darkness > 0, squash the density transformation towards 0 so that the
* slope near 0 is smaller by a factor of darkness
* If darkness < 0, squash the density transformation towards 1 so that the
* slope near 1 is smaller by a factor of |darkness|
* If darkness = 0, the density distribution is not changed.

      if(darkness.lt.0) then
          factor = -1 - 1/darkness
      else
          factor = darkness
      end if
      do 60 j = 1,ndens
60    dens(j) = dens(j)/(1+factor*(1-dens(j)))

      return
      end

      subroutine mgofatpixmap(ndens, dens, adjust, neg)
* Routine to adjust the density map to account for the size of most printer
* pixels. (A dark dot is larger than an equivalent light one)
* The coefficients d0 and d1 are empirically chosen so that a gray scale
* actually varies evenly from white to black.
      real dens(ndens)
      logical neg
      f(x) = 1 + (x-1)*(d1 + (x-1)*(d0+2*d1-3 + (x-1)*(d0+d1-2)))
      if(neg) then
          d0 = adjust
          d1 = 1/adjust
      else
          d0 = 1/adjust
          d1 = adjust
      end if
      do 900 i = 1,ndens
          dens(i) = f(dens(i))
900   continue
      return
      end

      subroutine mgoprdens(dmin,dmax,ndens,dens,count)
      include 'luns.par'
      real dens(1)
      integer count(1), total
      real step, buf(3,48)
      data step /.05/

      adens = ndens / (dmax-dmin)
      j1 = 1
      extra = 0
      n = nint(1/step)
      do 10 j = 1,n
      x = (j-1) * step
      do 900 i = 2,ndens
          j2 = i
          if(dens(i).gt.x) goto 12
900   continue
12    continue
      diff = dens(j2)-dens(j2-1)
      if(diff.eq.0) diff = 1
      frac = (x-dens(j2-1)) / diff
      xcount = j2*frac + (j2-1)*(1-frac)
      y = xcount/adens + dmin
      total = extra
      do 901 i = j1,j2-1
          total = total + count(i)
901   continue
      total = total + frac*count(j2)
      extra = (1-frac)*count(j2)
      j1 = j2 + 1
      buf(1,j) = x
      buf(2,j) = y
      buf(3,j) = total
10    continue
      write(stdout,1001)
1001  format(1x,2('   Density      Data    # pix ',5x))
      write(stdout,1000) 
     1	((buf(1,j),buf(2,j),nint(buf(3,j)),j=k,n,n/2),k=1,n/2)
1000  format(24(1x,2f10.3,i9,6x,2f10.3,i9/))
      return
      end
