CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
CC                  Mongo Interactive Graphics Software                    CC
CC                 Copyright (c) 1987, 1994 - John Tonry.                  CC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine mgospline(x,y,nknot)
C Calculates a natural spline from NKNOT points in X and Y
C Can then be evaluated by "set y u spline"
C NOTE: arrays x and y are sorted by x
C       maximum number of knot points is 100
      parameter (maxknot=100)
      real x(nknot), y(nknot)
      real yt(maxknot), y2(maxknot), h(maxknot)
      real a(maxknot), b(maxknot), c(maxknot), l(maxknot)
      real f(maxknot), r(maxknot), u(maxknot)

      common /mgosplinev/ npt, a, yt, y2, h

      if (nknot .gt. maxknot) then
         write(6,*) 'The number of spline points must be <=',maxknot
         return
      end if
      npt = nknot

* Sort the arrays
      do 110 j = nknot-1,1,-1
         k = j
         do 120 i = j+1,nknot
            if(x(j).le.x(i)) goto 121
            k = i
 120     continue
 121     continue
         if(k.eq.j) goto 110
         tempx = x(j)
         tempy = y(j)
         do 130 i = j+1,k
            x(i-1) = x(i)
            y(i-1) = y(i)
 130     continue
         x(k) = tempx
         y(k) = tempy
 110  continue

C     Generate spline equations
      n  = nknot - 1

      do 2758 i= 1, n
         h(i) = x(i+1) - x(i)
         if(h(i).eq.0) then
            write(6,*) 'Must have distinct spline knots'
            return
         end if
         f(i) = y(i+1) - y(i)
         r(i) = f(i)/h(i)
 2758 continue

      na = n - 1
      do 2759 i= 1, na
         b(i)= 6.0 * (r(i+1)-r(i))
         c(i)= h(i+1)
         a(i)= 2.0 * (h(i+1)+h(i))
 2759 continue

C Solve for second derivatives (set to zero at endpoints)
      y2(1) = 0.0
      y2(nknot) = 0.0
      u(1) = a(1)
      do 2760 i= 2, na
         l(i) = c(i-1)/u(i-1)
         u(i) = a(i) - l(i)*c(i-1)
 2760 continue

      yt(1) = b(1)
      do 2761 i= 2, na
         yt(i) = b(i) - l(i)*yt(i-1)
 2761 continue

      y2(na+1)= yt(na)/u(na)
      na1 = na - 1
      do 2762 i= 1, na1
         fac = yt(na-i) - c(na-i)*y2(na+2-i)
         y2(na+1-i) = fac/u(na-i)
 2762 continue

* Also fill the A and YT array with the original knots for later evaluation
      do 2763 i= 1, nknot
         y2(i) = y2(i)/6.0
         a(i) = x(i)
         yt(i) = y(i)
 2763 continue

      return
      end

      subroutine mgosplineval(xp,yp)
      parameter (maxknot= 100)
      real x(maxknot), y(maxknot), y2(maxknot), h(maxknot)
      common /mgosplinev/ npt, x, y, y2, h

C     Find the value of the spline at XP
      loc = 0
      do 2764 i= 1, npt
         if (xp .gt. x(i)) loc = loc + 1
 2764 continue

      if (loc .gt. 0 .and. loc .lt. npt) then
         sp1 = y2(loc)/h(loc)
         sp2 = y2(loc+1)/h(loc)
         sp3 = y(loc+1)/h(loc) - y2(loc+1)*h(loc)
         sp4 = y(loc)/h(loc) - y2(loc)*h(loc)
         dxp = x(loc+1) - xp
         dxm = xp - x(loc)
         yp = sp1*dxp**3 + sp2*dxm**3 + sp4*dxp + sp3*dxm

      else if (loc .eq. 0) then
C     If values are needed beyond the knot points, extrapolate with
C     the slope values on either end.
         deriv = -3.0*y2(1)/h(1)*(x(2) - x(1))**2.0
         deriv = deriv + (y(2)-y(1))/h(1) + (y2(1)-y2(2))*h(1)
         del = xp - x(1)
         yp = del*deriv + y(1)
      else
         deriv = (y(npt)-y(npt-1))/h(npt-1)+(y2(npt-1)-y2(npt))*h(npt-1)
         deriv = deriv + 3.0*y2(npt)/h(npt-1)*(x(npt)-x(npt-1))**2.0
         del = xp - x(npt)
         yp = del*deriv + y(npt)
      end if
      return
      end
