CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
CC                  Mongo Interactive Graphics Software                    CC
CC                 Copyright (c) 1987, 1994 - John Tonry.                  CC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
* Program to send a MONGO vector file to a QMS Lasergrafix laser printer
      parameter (nin=2048, nbuf=1048576, laserwidth=2550)
      integer*2 sign, buf(nbuf), m1, m2, ix0, iy0, cbuf(4*laserwidth)
      integer px2, py2
      character*132 line
      character*7 ncopy
      data ncopy /'^DC0001'/
      data sign /-32768/
      data iunit /6/

1000  format(a)
1001  format(a,i4.4,':',i4.4,a,i4.4,':',i4.4)
1002  format(a,i2.2)
1003  format(1x,a,i1)

******************************************************************************
* Vector file contains pairs of 16 bit numbers according to:                 *
*    Sign 1   Sign 2   Mantissa 1   Mantissa 2   Meaning                     *
*      0        0          x            y        Draw a dot at (x,y)         *
*      1        0          x            y        Relocate at (x,y)           *
*      0        1          x            y        Draw to (x,y)               *
*      1        1        32767        32767      End of vector file          *
*      1        1          0            0        Finish plot, start new page *
*      1        1          1            n        Line weight = n             *
*      1        1          2            n        Line type = n               *
* (x,y) are oriented in portrait mode                                        *
******************************************************************************

      px2 = laserwidth
      py2 = nint(10.666667/8.5*laserwidth)

* Read in list of vectors and find how many there are
      read(5,1000) line
      open (unit=1,file=line,form='unformatted',status='old')
      ncoord = 0
      do 1 j = 1,nbuf/nin
          read(1,end=2) (buf(i+ncoord),i=1,nin)
          ncoord = ncoord + nin
1     continue
2     continue

* Initialize the laser printer
*        Turn on QUIC
*     // Ask for NCOPY copies
*     // Ask for landscape orientation (requiring x<->y)
        line = '^PY^-' // ncopy // '^IOL'
        write(iunit,1000) line(:nch(line))
* Initialize syntax to default with dimensions in dots
        line = '^ISYNTAX00010'
        write(iunit,1000) line(:nch(line))
* Initialize tray and stacker to default
        line = '^IFXX0X'
        write(iunit,1000) line(:nch(line))
*     // Initialize top margin
*     // Initialize left origin
*     // Turn on Free Format
*     // Turn on Vector Graphics mode
*     // Set pen width
        line = '^IJ0' // '^IT0' // '^F^-' // '^IGV' // '^PW03'
        write(iunit,1000) line(:nch(line))

* Run through the vectors, sending them to the printer
* Note that several passes may be necessary for multipage plots
      ltype = 0
      maxx = 0
20    continue
      minx = maxx
      maxx = maxx + laserwidth
      n = 0
      more = 0
      do 10 i = 1,ncoord,2
          i1 = and(sign,buf(i))
          m1 = buf(i) - i1
          i2 = and(sign,buf(i+1))
          m2 = buf(i+1) - i2
* Test to see whether the vector is on the present page
          if(i1.eq.0.or.i2.eq.0) then
              if(m1.gt.maxx) then
                  more = 1
                  goto 10
              else if(m1.lt.minx) then
                  goto 10
              end if
          end if

* If the current operation can exceed the line length, write the line
          if(n.gt.132-22) then
              write(iunit,1000) line(:n)
              n = 0
          end if
* point or draw?
          if(i1.eq.0) then
              m1 = m1 - minx
              if(i2.eq.0) then
* point at m1, m2
                  write(line(n+1:n+22),1001) '^U', m2, m1,'^D', m2, m1
                  n = n + 22
              else
* draw to m1, m2
* If LTYPE = 0 (solid line) no problem; otherwise call chopper
                  if(ltype.eq.0) then
                      write(line(n+1:n+11),1001) '^D', m2, m1
                      n = n + 11
                      goto 10
                  end if
                  call chopper(ix0,iy0,m1,m2,ltype,1,px2,py2,cbuf,nv)
                  do 15 j = 1,nv,4
                      if(n.gt.132-22) then
                          write(iunit,1000) line(:n)
                          n = 0
                      end if
                      write(line(n+1:n+22),1001) 
     1				'^U', cbuf(j+1), cbuf(j),  
     1				'^D', cbuf(j+3),cbuf(j+2)
                      n = n + 22
15                continue
                  ix0 = m1
                  iy0 = m2
              end if
* relocate or special?
          else
* relocate at m1, m2
              if(i2.eq.0) then
                  ix0 = m1 - minx
                  iy0 = m2
                  write(line(n+1:n+11),1001) '^U', iy0, ix0
                  n = n + 11
* end of buffer
              else if(m1.eq.32767) then
                  goto 11
* eject page, start new one
              else if(m1.eq.0) then
                  if(n.gt.0) write(iunit,1000) line(:n)
                  n = 0
                  write(iunit,1000) '^,'
* set line weight
              else if(m1.eq.1 .and. m2.le.20) then
                  write(line(n+1:n+5),1002) '^PW',2*m2+1
                  n = n + 5
* set line type
              else if(m1.eq.2.and.m2.ge.0.and.m2.le.6) then
                  ltype = m2
              end if
          end if
10    continue

11    continue
      if(more.eq.0) goto 999
      if(n.gt.0) write(iunit,1000) line(:n)
      write(iunit,1000) '^,'
      goto 20

999   continue
* Write the LINE buffer if it is not empty and print the page
      if(n.gt.0) write(iunit,1000) line(:n)
* Apparently this last form feed is not necessary. 
C      write(iunit,1000) '^,'
* Finish up the plot
* Turn off Vector Graphics mode
*     // Turn off Free Format
*     // Turn off QUIC
*     // Reset QUIC to default mode
      line = '^U0000:0000^ISYNTAX00000^IGE' // '^O' // '^-^PN^-'
      write(iunit,1000) line(:nch(line))

      stop
      end

      function nch(s)
      character*(*) s
      do 900 i = len(s),1,-1
          nch = i
          if(s(i:i).ne.' '.and.ichar(s(i:i)).ne.0) return
900   continue
      return
      end
