/*CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
CC                  Mongo Interactive Graphics Software                    CC
CC                 Copyright (c) 1987, 1994 - John Tonry.                  CC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC*/
#include <stdio.h>
#include <string.h>

#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define ABS(a) (((a) > 0) ? (a) : -(a))

/* #define TEST */

/* mgo_gets will try to fork off a process which can use readline() to
 * process input, returning it line by line.  Failing that, mgo_gets will
 * default to plain old gets().  Use mgo_gets_error=1 to disable fork...
 */
static int mychildpid = (-1);
static int mgo_gets_error=0, handshook=0;
static char prompt[20];

static int mgoinfd=0, mgoutfd;
static int mgxdtwidth=(-1);           /* Descriptor table width */

/* Define a table of descriptors and handlers */
#define MGOXIN_PUT_PROMPT     1
#define MGOXIN_RETURN         2

#define MAXSELECT 8
struct select_struct {
  int infd;		/* input descriptor */
  int outfd;		/* output descriptor */
  char * (*infunc)();	/* input handler */
  char * (*outfunc)();	/* output (prompt) handler */
  int flags;		/* flags */
};

static struct select_struct inputs[MAXSELECT]={
    -1,-1,NULL,NULL,0,
    -1,-1,NULL,NULL,0,
    -1,-1,NULL,NULL,0,
    -1,-1,NULL,NULL,0,
    -1,-1,NULL,NULL,0,
    -1,-1,NULL,NULL,0,
    -1,-1,NULL,NULL,0,
    -1,-1,NULL,NULL,0
};



mgo_gets_init(fdin, fdout)
int *fdin, *fdout;
{
  int pfdin[2], pfdout[2];
  char *chargv[6], chargs[6][32];
  __pid_t fork();
  ssize_t write(); 
  void _exit();
 
/* Open pipes and fork */
  if( pipe(pfdin) == -1 || pipe(pfdout) == -1) {
    fprintf(stderr, "can't open pipes\n");
    return 1;
  }

  switch( (mychildpid=fork()) ) {
 
  case -1:
    fprintf(stderr,"can't fork\n");
    return 2;
 
  case 0:
 
/* Create args for process we are about to exec */
    strcpy(chargs[0], "mongoin");         chargv[0] = chargs[0];
    strcpy(chargs[1], "-in");             chargv[1] = chargs[1];
    sprintf(chargs[2], "%d", pfdout[0]);  chargv[2] = chargs[2];
    strcpy(chargs[3], "-out");            chargv[3] = chargs[3];
    sprintf(chargs[4], "%d", pfdin[1]);   chargv[4] = chargs[4];
    chargv[5] = NULL;
 
/* Close the unused child descriptors */
    if(close(pfdout[1]) == -1 || close(pfdin[0]) == -1) {
      fprintf(stderr,"Error closing child's pipes\n");
      return(3);
    }

/* Go off and exec mongoin */
#ifdef TEST
    fprintf(stderr, "About to exec\n");
#endif
    execvp("mongoin", chargv);

#ifdef TEST
    fprintf(stderr, "Failed to exec 'mongoin'\n");
#endif

/* It failed, time to tell the parent and prepare to die... */
    write(pfdin[1], "child", 6);
    return(4);

  default:
    ;
  }
 
/* Meanwhile back at the parent */
 
/* Close the unused parent descriptors */
  if(close(pfdout[0]) == -1 || close(pfdin[1]) == -1) {
    fprintf(stderr,"Error closing parent's pipes\n");
    return(5);
  }
 
/* and tell us the ones which communicate with the child */
  *fdin = pfdin[0];
  *fdout = pfdout[1];

  return(0);
}



char *mgo_gets(string, len)
char *string;
int len;
{
  int nread;
  char *c = string;
  ssize_t read(); 
 
/* Read a line and return */

  if(mgo_gets_error == 0) {
 
/* See what the child has for us */
#ifdef TEST
    fprintf(stderr,"Parent: waiting to read %d into %d...\n", len, string);
#endif
    read(mgoinfd, string, len);
    nread = strlen(string);
 
/* EOF? */
    if(string[0] == EOF) {
#ifdef TEST
      fprintf(stderr, "parent EOF, quit time\n");
#endif
      kill(mychildpid, 9);
      return(NULL);
    }
 
/* Otherwise fall back on normal gets() */
  } else {
#ifdef TEST
    fprintf(stderr, "parent reading with gets, err = %d\n", mgo_gets_error);
#endif
    
    c = fgets(string, len, stdin);
    nread = strlen(string);
  }

/* Terminate the line and return */
  if(string[nread-1] == '\n' || string[nread-1] == '\0') {
    string[nread-1] = '\0';
  } else {
    string[nread] = '\0';
  }
#ifdef TEST
    fprintf(stderr, "parent (%2d): '%s'\n", nread, string);
#endif

  return(c);
}


mgoxinprompt(fd, prompt)
int fd;
char *prompt;
{
  ssize_t write(); 
  write(fd, prompt, strlen(prompt));
  return(0);
}


int mgoxinclose_()
{
  kill(mychildpid, 9);
  return(0);
} 

mgoxinopen_()
{
  int nread;
  char buf[20];
  ssize_t read(); 
/* If the first time, try to fork and exec 'mongoin' to run readline() */
  mgo_gets_error = mgo_gets_init(&mgoinfd, &mgoutfd);
#ifdef TEST
    fprintf(stderr,"Parent: err = %d, in/out = %d %d\n", mgo_gets_error, mgoinfd, mgoutfd);
#endif

/* working on trying to kill the child if the exec fails... */
  if(mychildpid > 0 && mgo_gets_error == 0 && handshook == 0) {
    read(mgoinfd, buf, 20);
    nread = strlen(buf);
#ifdef TEST
    fprintf(stderr, "Parent: got handshake from %s (%d)\n", buf, nread);
#endif
    if(strncmp(buf, "child", 5) == 0) {
      fprintf(stderr,"Failed to exec input editing routine 'mongoin',");
      fprintf(stderr," will use straight input...\n");
      close(mgoinfd);
      close(mgoutfd);
      kill(mychildpid, 9);
      mgoinfd = 0;
      mgo_gets_error = 2;
      mgoinfd = 0;
      mgoutfd = 1;
    }
    handshook = 1;
  }

  mgoaddfd(mgoinfd, mgoutfd, mgo_gets, mgoxinprompt,
	   MGOXIN_PUT_PROMPT | MGOXIN_RETURN);

  return(mgo_gets_error);
}

mgoaddfd(infd, outfd, inhandler, outhandler, flags)
     int infd;
     int outfd;
     char * (*inhandler)();
     char * (*outhandler)();
     int flags;
{
  int i;
  for(i=0; i<MAXSELECT; i++) {
    if(inputs[i].infd == -1) {
      inputs[i].infd = infd;
      inputs[i].outfd = outfd;
      inputs[i].infunc = inhandler;
      inputs[i].outfunc = outhandler;
      inputs[i].flags = flags;

#ifdef TEST
      fprintf(stderr,"Added handler at %d, fd's = %d and %d, flags = %d\n",
	      i, infd, outfd, flags);
#endif

      return(0);
    }
  }
  return(1);
}

mgodelfd(fd)
     int fd;
{
  int i, j;
  for(i=0; i<MAXSELECT; i++) {
    if(inputs[i].infd == fd) {
      for(j=i; j<MAXSELECT-1; j++) {
	inputs[j].infd = inputs[j+1].infd;
	inputs[j].outfd = inputs[j+1].outfd;
	inputs[j].infunc = inputs[j+1].infunc;
	inputs[j].outfunc = inputs[j+1].outfunc;
	inputs[j].flags = inputs[j+1].flags;
      }
      inputs[MAXSELECT-1].infd = -1;
      inputs[MAXSELECT-1].infunc = NULL;
      return(0);
    }
  }
  return(1);
}

/* This mgoxin uses select to service X events and read from stdin */
#include <X11/Xos.h>

int mgoxin_(pstring, string, plen, len)
char *pstring, *string;
int plen, len;
{
  int i, j;
  fd_set readfds;
  char *c;

  int nread;

  if(len > 4096) {
    fprintf(stderr,"WHOA! mgoxin: len = %d",len); /* Sanity */
    return(0);
  }

  if(mgxdtwidth < 0) mgxdtwidth = getdtablesize();

/* Pad the prompt string with spaces */
  while (--plen >= 0 && pstring[plen] == ' ') ;
  strncpy(prompt+1, pstring, plen+1);
  prompt[0] = prompt[plen+2] = ' ';
  prompt[plen+3] = '\0';
#ifdef TEST
  fprintf(stderr,"Setting prompt to '%s' (%s) pid = %d, err = %d\n", prompt, pstring, mychildpid, mgo_gets_error);
#endif

/* Put out a prompt onto those fd's requesting it */
  for(i=0; i<MAXSELECT; i++) {
    if(inputs[i].infd >= 0 && inputs[i].outfunc != NULL && 
       inputs[i].flags & MGOXIN_PUT_PROMPT) {
      (*(inputs[i].outfunc))(inputs[i].outfd, prompt);
      
#ifdef TEST
  fprintf(stderr,"Putting out prompt for %d to %d\n", i, inputs[i].outfd);
#endif
    }
  }

/* Loop until a line of input appears (while handling other events) */
  while(1) {
    FD_ZERO(&readfds);
    for(i=0; i<MAXSELECT; i++) {
/* Watch active fd's */
      if(inputs[i].infd >= 0) FD_SET(inputs[i].infd, &readfds);
    }
    i = -1;
/* Continue if select returns a negative number (signal) */
    while (i < 0) {
      i = select(mgxdtwidth, &readfds, (fd_set *)NULL, (fd_set *)NULL,
		 (struct timeval *)NULL);
    }

/* Handle the various descriptors which have something for us */
    for(i=0; i<MAXSELECT; i++) {
      if(inputs[i].infd >= 0 && FD_ISSET(inputs[i].infd, &readfds)) {
	c = (*(inputs[i].infunc))(string, len);
	if(inputs[i].flags & MGOXIN_RETURN) {
#ifdef TEST
	  fprintf(stderr,"i = %d, c = %d, string = '%s'\n", i, c, string);
#endif
/* Fill the string to the end with spaces for Fortran */
	  for(j=strlen(string); j<len; j++) string[j] = ' ';
	  if (c == NULL) return(0);       /* Return = 0/1 for EOF/line */
	  else return(1);
	}
      }
    }
  }
}
