/*CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
CC                  Mongo Interactive Graphics Software                    CC
CC                 Copyright (c) 1987, 1994 - John Tonry.                  CC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC*/
#include <stdio.h>

#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define ABS(a) (((a) > 0) ? (a) : -(a))

int width, height, minify;	       /* last size of window */
int waiting_for_cursor;		       /* flag for when cursor is on */
int last_coord_point;
int vclass;			       /* Visual class */

/**************************************************************/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#define DEFAULT_FONT "9x15"     /* default font name */
#define MAXCOLORS 32            /* maximum number of colors requested */
#define LINE_CAP CapRound       /* End of lines (could use CapProjecting) */
#define LINE_JOIN JoinRound     /* Line joints (cound use JoinBevel) */
#define MAXBUF 200		/* maximum number of buffer pointers saved */
#define MAXVEC 1024		/* maximum number of vectors per buffer */
#define MAXSTRING 256		/* maximum number of strings saved */
#define MAXSBUF 10240		/* maximum length of strings saved */

/* There seems to be a bug in some X servers with line_weight > 0, whereby
 * lines drawn in the graphics window will show up in overlaying windows.
 * Ameliorate its effects by replacing "1" with "0" */
#define LWEIGHT (lweight == 1 ? 0 : lweight)

/* #define SEEREPLAY */

char *malloc();

/* Define the various windows */
Display *mgxdpy=NULL;                      /* Display pointer */
Window mgxwin = (Window)0 ;		    /* Base window */
int mgxscrn;                               /* Mgxscrn number of display */
XSetWindowAttributes mgxswa;               /* Window info structure */
GC mgxGC;                                  /* The graphics context */
Window mgxinfowin;                         /* For retrieving info */
int mgxinfox, mgxinfoy, mgxinfow, mgxinfoh, mgxinfob, mgxinfod;  /* ditto */

/* Define the window geometry arrays */
char *mgxdisplay = NULL;

/* Define the font used for the xyz display */
Font mgxfont;
char *mgxfontname=DEFAULT_FONT;
XFontStruct *mgxfontinfo;
int mgxfontheight, mgxfontwidth;

Cursor mgxcross;                       /* Cursor for cursor mode */

int mgxnclr;				/* number of vector colors used */
unsigned long mgxpx[256];		/* color cell addresses allocated */
unsigned long mgxback, mgxfore;	/* back/foreground colors */
Colormap mgxcmap;			/* Display color map */
XColor mgxcolor[MAXCOLORS];		/* vector color palette */

int color, lastcolor;                   /* Color for current operation */
int lweight;                            /* Line weight */
int ltype;                              /* Line type */
int fillstart, fillend;                 /* Range of fill sequence */

int MGXK=1;                            /* Is it safe to service X events? */
                                        /* (This is for an interrupt driven */
                                        /* X handler.  It is harmless for a */
                                        /* select driven X handler) */
int REPLAY=0;                           /* Are we in the midst of a redraw? */
int FILLING=0;                          /* Are we tracing an area to fill? */
int FILLMODE=0;                         /* Type of filling to do */
int FILLNEW=0;                          /* Brand new fill command? */
int RELOCMODE=0;                        /* Fill at relocate? */

#define MAXFILLPTS 1024
XPoint mgxfillpts[MAXFILLPTS];

struct mgxsavebuf {			/* Define vector save structure */
  unsigned short w1;			/* first word */
  unsigned short w2;			/* second word */
};

struct mgxsavebuf *savebuf[MAXBUF];    /* Pointers to vector save buffers */
char stringbuf[MAXSBUF];                /* Buffer for saved strings */
int nsave=0;	                        /* x,y of last vector; number drawn */
int stringptr=0;                        /* String pointer within stringbuf */

/* Codes for how things are saved */
/* The signs of word 1 (1's bit) and word 2 (2's bit) */
#define DOT_SAVE  0    /* This is a dot, mantissas are x,y */
#define RELO_SAVE 1    /* This is a relocate, mantissas are x,y */
#define DRAW_SAVE 2    /* This is a draw, mantissas are x,y */
#define ETC_SAVE  3    /* This is something else, see below */
/* The first mantissa m1, and second mantissa m2 */
#define END_SAVE  32767
#define FINI_SAVE 0    /* Finish plot, m2 = 0*/
#define LW_SAVE   1    /* Line weight change, m2 = lweight */
#define LT_SAVE   2    /* Line type change, m2 = ltype */
#define CLR_SAVE  3    /* Palette entry change, m2 = color */
#define RED_SAVE  4    /* Set red entry of current palette, m2 = red (0-255) */
#define GRN_SAVE  5    /* Set green entry of current palette, m2 = green */
#define BLUE_SAVE 6    /* Set blue entry of current palette, m2 = blue */
#define STR_SAVE  7    /* Draw string, m2 = ptr to string buffer */
#define FILL_SAVE 8    /* Change fill mode, m2 = fill type + 1024*RELOCMODE */

int xlast, ylast;			/* x,y of last vector */

int mousex, mousey;			/* loc where button/key went down */
float mousex0, mousedx, mousey0, mousedy;  /* mouse -> user coord transform */

long int mgxeventmask;			/* Those events we will service */
char *mgxrefresh();			/* Interrupt service routine */

#define ERR_CANT_OPEN_DISPLAY			1
#define ERR_CANT_CREATE_IMAGE_WINDOW		2
#define ERR_CANT_CREATE_ZOOM_WINDOW		3
#define ERR_INSUFF_VECTOR_COLORS		4
#define ERR_INSUFF_STATUS_COLORS		5
#define ERR_INSUFF_IMAGE_COLORS			6
#define ERR_BAD_ALLOC_LOOKUP			7
#define ERR_BAD_ALLOC_PALETTE			8
#define ERR_BAD_ALLOC_IMBUF			9
#define ERR_BAD_ALLOC_ZOOM			10
#define ERR_NO_DATA				11
#define ERR_BAD_DISPLAY_PIPE			12
#define ERR_NOT_INITIALIZED			13
#define ERR_CANT_GET_FONT			14


#define MAXSTIPPLE 19
#define STIPPLE_SIZE 8
static unsigned char stipple_bits[MAXSTIPPLE][STIPPLE_SIZE] = {
  0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa,  /* gray 80 */
  0x55, 0x00, 0x55, 0x00, 0x55, 0x00, 0x55, 0x00,  /* gray 50 */
  0x11, 0x00, 0x00, 0x00, 0x11, 0x00, 0x00, 0x00,  /* gray 20 */
  0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11,  /* vertical stripes */
  0xff, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00,  /* horizontal stripes */
  0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04,  /* low dens vertical */
  0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00,  /* low dens horizontal */
  0x22, 0x44, 0x88, 0x11, 0x22, 0x44, 0x88, 0x11,  /* left slant stripes */
  0x44, 0x22, 0x11, 0x88, 0x44, 0x22, 0x11, 0x88,  /* right slant stripes */
  0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80,  /* low dens left slant */
  0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01,  /* low dens right slant */
  0xff, 0x11, 0x11, 0x11, 0xff, 0x11, 0x11, 0x11,  /* grid */
  0x04, 0x04, 0xff, 0x04, 0x04, 0x04, 0x04, 0x04,  /* low dens grid */
  0x81, 0x42, 0x24, 0x18, 0x18, 0x24, 0x42, 0x81,  /* crisscross */
  0x42, 0x81, 0x18, 0x24, 0x24, 0x18, 0x81, 0x42,  /* tiny bubbles */
  0x00, 0x1c, 0x22, 0x22, 0x22, 0x1c, 0x00, 0x00,  /* big bubbles */
  0x24, 0x42, 0x81, 0x18, 0x18, 0x81, 0x42, 0x24,  /* bubbles and dots */
  0x81, 0x4e, 0x48, 0x78, 0x1e, 0x12, 0x72, 0x81,  /* Escher-like hatching */
  0x88, 0x1d, 0x3e, 0x5c, 0x88, 0xc5, 0xe3, 0xd1}; /* basket weave */
Pixmap stipmap[MAXSTIPPLE];
/*  0x08, 0x08, 0x14, 0xe3, 0x80, 0x80, 0x41, 0x3e}; /* fish scale */ 

/* Initialize X server connection and windows
 */

mgxinit(winit,hinit,minit,c,npal,palette)
int *winit, *hinit;	/* Size of window (request on input, actual returned */
int *minit;		/* Minification (request on input, actual returned */
int *c;			/* Initial color to use */
int *npal;		/* Number of color cells requested and actual */
short int *palette;     /* palette entries: [npal][3] */
{
  int i;
  long int planes;
  char *xdefstring;
  XColor colorreturn;
  int owncmap;
  Visual *visual;
  int vdepth;

/* Resize window and return if connection to the server is already open */
  if( mgxdpy != NULL) {
    minify = *minit;
    i = minify == 0 ? 1 : ABS(minify);
    width = *winit/i;
    height = *hinit/i;
    XResizeWindow(mgxdpy,mgxwin,width,height);
    *npal = mgxnclr;
    return(0);
  }
  
  owncmap = 0;
  mgxscrn = -1;
  vclass = -1;
  mgxnclr = *npal - 2;	   /* Since mgxopen() will give us fg and bg anyway*/
  minify = *minit;
  i = minify == 0 ? 1 : ABS(minify);
  width = *winit/i;
  height = *hinit/i;

  if(mgxopen(mgxdisplay, "xmongo", &mgxdpy, &mgxscrn, &mgxcmap, &owncmap, 
	   &mgxnclr, mgxpx, &mgxback, &mgxfore, 
	   &visual, &vclass, &vdepth, 
	   &width, &height, &mgxwin, &mgxswa) != 0) return(1);

/* Set the color addresses in mgxcolor to the addresses from the server */
  mgxnclr += 2;
  *npal = mgxnclr;

  for(i=0; i<mgxnclr; i++) {
    if(i == 0) {
/* background (xmongo*background) */
      colorreturn.pixel = mgxback;
      XQueryColor(mgxdpy, mgxcmap, &colorreturn);
    } else if(i == 1) {
/* foreground (xmongo*foreground) */
      colorreturn.pixel = mgxfore;
      XQueryColor(mgxdpy, mgxcmap, &colorreturn);
    } else {
      colorreturn.pixel = mgxpx[i-2];
    }
    mgxcolor[i].pixel = colorreturn.pixel;
    mgxcolor[i].red =   colorreturn.red;
    mgxcolor[i].green = colorreturn.green;
    mgxcolor[i].blue =  colorreturn.blue;
    mgxcolor[i].flags = DoRed | DoGreen | DoBlue;
/* Use just the first two colors for Mongo's color palette */
    if(i <= 1) {
      *(palette+3*(i)+0) = colorreturn.red / 256;
      *(palette+3*(i)+1) = colorreturn.green / 256;
      *(palette+3*(i)+2) = colorreturn.blue / 256;
    }
  }
  mgxchangeclrs(mgxdpy, mgxcmap, mgxcolor, mgxnclr);

/* Get the font to be used, and get the information about it */
  mgxfont = XLoadFont(mgxdpy,mgxfontname);
  mgxfontinfo = XQueryFont(mgxdpy,mgxfont);
  if(mgxfontinfo == NULL) {
    fprintf(stderr,"Can't open font %s",mgxfontname);
    return(ERR_CANT_GET_FONT);
  }
  mgxfontwidth = mgxfontinfo->max_bounds.rbearing - mgxfontinfo->min_bounds.lbearing;
  mgxfontheight = mgxfontinfo->max_bounds.ascent + mgxfontinfo->max_bounds.descent;

/* Load up the two cursors we might want to use */
  mgxcross = XCreateFontCursor(mgxdpy, XC_crosshair);

/* Create the graphics context for line drawing */
  mgxGC = XCreateGC(mgxdpy, mgxwin, (long)0, NULL);
  XSetState(mgxdpy, mgxGC, mgxcolor[1].pixel,
	    mgxcolor[0].pixel, GXcopy, AllPlanes);
  XSetLineAttributes(mgxdpy, mgxGC, 0, LineSolid, LINE_CAP, LINE_JOIN);
  XSetFillRule(mgxdpy, mgxGC, WindingRule); 

  XSetFont(mgxdpy, mgxGC, mgxfont);
  XSetFillStyle(mgxdpy, mgxGC, FillSolid);

/* Create stipple bit maps */
  for(i=0; i<MAXSTIPPLE; i++) {
    stipmap[i] = XCreateBitmapFromData(mgxdpy, mgxwin, stipple_bits[i], 
				       STIPPLE_SIZE, STIPPLE_SIZE);
  }

/* Select inputs from the window */
  mgxeventmask = ExposureMask | ButtonPressMask | ButtonReleaseMask |
               PointerMotionMask | PointerMotionHintMask | KeyPressMask;

  XSelectInput(mgxdpy, mgxwin, mgxeventmask);

/* Flush this pile of X output */
  XFlush(mgxdpy);

/* Tell mgoxin() that X is being used and which fd it is using */
  mgoaddfd(ConnectionNumber(mgxdpy), -1, mgxrefresh, NULL, 0);

  /* Finally, initialize a bunch of things */
  i = minify == 0 ? 1 : ABS(minify);
  *winit = width * i;
  *hinit = height * i;
  *minit = minify;

/* Initialize global mode variables */
  MGXK = 1;
  REPLAY = 0;
  FILLING = 0;

/* Initialize the save buffer pointers */
  for(i=0; i<MAXBUF; i++) savebuf[i] = NULL;
  nsave = 0;
  stringptr = 0;

/* Initialize line attributes */
  lastcolor = -1;
  lweight = 1;
  if(*c >= 0) mgxselcol(*c);
  else mgxselcol(1);
  mgxltype(0);
/* It seems there is a bug in some X servers with line weight 1.  
 * Hence this funny exercise with lweight = 0 */
  mgxlweight(0);
  mgxlweight(1);
}


mgxerase()
{
  nsave = 0;
  stringptr = 0;
  FILLING = 0;
  XClearWindow(mgxdpy,mgxwin);
/* Initialize the replay buffer with the current state */
  mgxselcol(color);
  mgxltype(ltype);
  mgxlweight(lweight);
  XFlush(mgxdpy);
}

mgxpalette(n,r,g,b)
int n;                  /* Number of table entries      */
short *r, *g, *b;       /* Pointer to red, green and blue arrays */
{
  int i, orig;
  orig = color;
  if(FILLING) mgxfill();

  for(i=0; i<MIN(mgxnclr,n); i++) {
    mgxcolor[i].red =   256 * (r[i] & 0x00ff);
    mgxcolor[i].green = 256 * (g[i] & 0x00ff);
    mgxcolor[i].blue =  256 * (b[i] & 0x00ff);
    
    if(!REPLAY) mgxsaveit(ETC_SAVE, CLR_SAVE, i);
    if(!REPLAY) mgxsaveit(ETC_SAVE, RED_SAVE,  (int)r[i]);
    if(!REPLAY) mgxsaveit(ETC_SAVE, GRN_SAVE,  (int)g[i]);
    if(!REPLAY) mgxsaveit(ETC_SAVE, BLUE_SAVE, (int)b[i]);
  }
  MGXK = 0;
  mgxchangeclrs(mgxdpy, mgxcmap, mgxcolor, mgxnclr);
  MGXK = 1;
  color = orig;
  return (0);
}


mgxselcol(c)
int c;
{
  color = c;
  if(FILLING) mgxfill();
  if(color != lastcolor && mgxnclr >= 2 && color < mgxnclr) {
    MGXK = 0;
    XSetForeground(mgxdpy, mgxGC, mgxcolor[color].pixel);
    lastcolor = color;
    XFlush(mgxdpy);
    MGXK = 1;
  }
  if(!REPLAY) mgxsaveit(ETC_SAVE, CLR_SAVE, c);
  return (0);
}

mgxpaladd(n,palette)
int n;
short int *palette;
{
  if(FILLING) mgxfill();
  if(n<0 || mgxnclr < 2 || n >= mgxnclr) return(-1);
  mgxcolor[n].red   = 256 * (palette[0] & 0x00ff);
  mgxcolor[n].green = 256 * (palette[1] & 0x00ff);
  mgxcolor[n].blue  = 256 * (palette[2] & 0x00ff);
  MGXK = 0;
  mgxchangeclrs(mgxdpy, mgxcmap, &mgxcolor[n], 1);
  MGXK = 1;
  if(!REPLAY) mgxsaveit(ETC_SAVE, CLR_SAVE, n);
  if(!REPLAY) mgxsaveit(ETC_SAVE, RED_SAVE,  (int)palette[0]);
  if(!REPLAY) mgxsaveit(ETC_SAVE, GRN_SAVE,  (int)palette[1]);
  if(!REPLAY) mgxsaveit(ETC_SAVE, BLUE_SAVE, (int)palette[2]);
  XFlush(mgxdpy);
  mgxselcol(n);
  return (0);
}

mgxlweight(w)
int w;
{
  if(FILLING) mgxfill();
  lweight = w;
  MGXK = 0;
  XSetLineAttributes(mgxdpy, mgxGC, LWEIGHT, 
		     (ltype==0) ? LineSolid:LineOnOffDash, 
		     LINE_CAP, LINE_JOIN);
  MGXK = 1;
  XFlush(mgxdpy);
  if(!REPLAY) mgxsaveit(ETC_SAVE, LW_SAVE, w);
  return (0);
}

char dash_list[6][4]={
  1, 3, 1, 3, 10, 7,10, 7, 20, 7,20, 7,
  10, 7, 1, 7, 20, 7, 1, 7, 20, 7,10, 7};
mgxltype(t)
int t;
{
  if(FILLING) mgxfill();
  MGXK = 0;
  if(t == 0) {
    ltype = 0;
  } else if(t <= 6) {
    ltype = t;
    XSetDashes(mgxdpy, mgxGC, 0, dash_list[ltype-1], 4);
  }
  XSetLineAttributes(mgxdpy, mgxGC, LWEIGHT, 
		     (ltype==0) ? LineSolid:LineOnOffDash, 
		     LINE_CAP, LINE_JOIN);
  MGXK = 1;
  XFlush(mgxdpy);
  if(!REPLAY) mgxsaveit(ETC_SAVE, LT_SAVE, t);
  return (0);
}

mgxrelocate(xuser, yuser)
int xuser, yuser;		/* x,y coordinates in user's system */
{
  if(!FILLNEW && xuser == xlast && yuser == ylast) return(0);
  FILLNEW = 0;
  if(!RELOCMODE && FILLING) mgxfill();
  if(FILLMODE > 0 && !FILLING) {
    FILLING = 1;
    fillstart = nsave;
  }
  xlast = xuser;
  ylast = yuser;
  if(!REPLAY) {
    mgxsaveit(RELO_SAVE, xuser, yuser);
  }
  return (0); 
}

mgxdraw(xuser, yuser)
int xuser, yuser;		/* x,y coordinates in user's system */
{
  int x, y, yorig;
  x = xuser;
  y = height-1 - yuser;
  yorig = height-1 - ylast;
  if(lweight > 0) {
    MGXK = 0;
    if(xuser != xlast || yuser != ylast) {
      XDrawLine(mgxdpy,mgxwin,mgxGC,xlast,yorig,x,y);
    } else {
      XDrawPoint(mgxdpy,mgxwin,mgxGC,x,y);
    }
    MGXK = 1;
  }
  xlast = xuser;
  ylast = yuser;

  if(!REPLAY) {
    mgxsaveit(DRAW_SAVE, xuser, yuser);
    fillend = nsave;
  }
  return (0); 
}

mgxwarp(x, y)
int x, y;
{
  if(FILLING) mgxfill();
  XWarpPointer(mgxdpy, None, mgxwin, 0,0,0,0, x, height-y-1);
}

mgxcursor(r, x, y, x0, dx, y0, dy)
int r, *x, *y;
float *x0, *dx, *y0, *dy;
{
  XEvent event;
  char key;
  if(FILLING) mgxfill();
  MGXK = 0;
  XDefineCursor(mgxdpy, mgxwin, mgxcross);
  MGXK = 1;
  waiting_for_cursor = r;		/* Set flag for interupt routine */
  last_coord_point = -1;
/* Set the local copies of the cursor scale variables */
  mousex0 = *x0;
  mousedx = *dx;
  mousey0 = *y0;
  mousedy = *dy;
  while(1) {
/* Block until X events are available; service them */
    XPeekEvent(mgxdpy, &event);
    mgxrefresh(0);
/* Return when key has been struck */
    if(waiting_for_cursor == 0) {
      *x = mousex;
      *y = mousey;
      MGXK = 0;
      XUndefineCursor(mgxdpy, mgxwin);
      MGXK = 1;
      return(0);
    }
  }
}

#include <math.h>

#define COORDLEN 200

mgxnewcoords(info,x,y)
int info, x, y;
{
  int textx, texty;
  float xuser, yuser;
  int imousex, imousey;
  double range;
  int nsigx, nsigy;
  char coordstring[COORDLEN], formatstring[32];
  int i, mini, nline, xylen, f;
  textx = texty = 20;
  f = minify == 0 ? 1 : ABS(minify);
  xuser = f*x*mousedx + mousex0;
  yuser = f*(height-1-y)*mousedy + mousey0;

  range = MAX(mousex0, mousex0+width*mousedx) / (width*mousedx);
  range = log10(MAX(0.01,ABS(range)));
  nsigx = MAX(3,(int)(range+4));
  nsigx = MIN(6, nsigx);
  range = MAX(mousey0, mousey0+height*mousedy) / (height*mousedy);
  range = log10(MAX(0.01,ABS(range)));
  nsigy = MAX(3,(int)(range+4));
  nsigy = MIN(6, nsigy);

  sprintf(formatstring,"(%%7.%dg,%%7.%dg)   ",nsigx,nsigy);
  sprintf(coordstring,formatstring,xuser,yuser);
  xylen = strlen(coordstring);

  XDrawImageString(mgxdpy, mgxwin, mgxGC, textx, texty,
		   coordstring, xylen);

  if(info != 0) {
    mini = 1;
    i = mgoinfoline_(&mini,&xuser,&yuser,&nline,coordstring,COORDLEN);
    if(i != last_coord_point) {
      last_coord_point = i;
      XDrawImageString(mgxdpy, mgxwin, mgxGC, 
		       textx+mgxfontwidth*(xylen+1), texty,
		       coordstring, nline+10);
    }
  }
}

mgxstring(xuser,yuser,string)
int xuser, yuser;
char *string;
{
  int x, y, len;
  if(FILLING) mgxfill();

  x = xuser;
  y = height-1 - yuser;

  len = MIN(255,strlen(string));
  MGXK = 0;
  XDrawImageString(mgxdpy,mgxwin,mgxGC,x,y,string,len);
  MGXK = 1;

  if(stringptr+len+1 < MAXSBUF) {         /* There is room to save it */
    if(!REPLAY) {
      mgxsaveit(RELO_SAVE, xuser, yuser);
      mgxsaveit(ETC_SAVE, STR_SAVE, stringptr);
      strncpy(stringbuf+stringptr, string, len);
      stringptr += len;
      stringbuf[stringptr++] = '\0';
    }
  }
  return(0);
}

mgxidle()
{
/*
  if(FILLING) mgxfill();
*/
/*  if(!RELOCMODE && FILLING) mgxfill(); */
  XFlush(mgxdpy);
}

mgxsetfill(f, r)
int f, r;
{
  if(FILLING) mgxfill();
/* f = 0 reserved for "off" and f = 1 reserved for "solid" */
  if(f < 0 || f >= MAXSTIPPLE+2) return(-1);
  FILLMODE = f;
  if(r == 0) RELOCMODE = 0;
  else RELOCMODE = 1;
  MGXK = 0;
  if(FILLMODE > 1) XSetStipple(mgxdpy, mgxGC, stipmap[f-2]);
  MGXK = 1;
  FILLNEW = 1;
  if(!REPLAY) mgxsaveit(ETC_SAVE, FILL_SAVE, FILLMODE+1024*RELOCMODE);
}

mgxfill()
{
  int n, i, buf;
  if(fillend-fillstart > 2) {
    MGXK = 0;
/*    if(FILLMODE > 1) XSetFillStyle(mgxdpy, mgxGC, FillStippled); */
/* Much as I like transparent fills, compatibility with Postscript
 * suggests that I stick to an opaque fill.  It has advantages, too... */
    if(FILLMODE > 1) XSetFillStyle(mgxdpy, mgxGC, FillOpaqueStippled);
    MGXK = 1;
    for(n=fillstart; n<MIN(fillend,fillstart+MAXFILLPTS); n++) {
      buf = n / MAXVEC;
      i = n % MAXVEC;
      mgxfillpts[n-fillstart].x = savebuf[buf][i].w1 & 0x7fff;
      mgxfillpts[n-fillstart].y = (height-1-savebuf[buf][i].w2) & 0x7fff;
/*
      fprintf(stderr,"Fill: %8d %8d\n",savebuf[buf][i].w1,savebuf[buf][i].w2);
*/
    }

    MGXK = 0;
    XFillPolygon(mgxdpy, mgxwin, mgxGC, mgxfillpts,
	       fillend-fillstart, Complex, CoordModeOrigin);
    if(FILLMODE > 1) XSetFillStyle(mgxdpy, mgxGC, FillSolid);
    MGXK = 1;
  }
  FILLING = 0;
}

mgxreplay()
{
  static int c;
  int i, n, m1, m2, signs, buf, NSAVE;
  short int pal[3];

  REPLAY = 1;
  NSAVE = nsave;

  for(nsave=0; nsave<NSAVE; nsave++) {
    buf = nsave / MAXVEC;
    i = nsave % MAXVEC;
    signs = ((savebuf[buf][i].w1 & 0x8000)>>15) + 
      ((savebuf[buf][i].w2 & 0x8000)>>14);
    m1 = savebuf[buf][i].w1 & 0x7fff;
    m2 = savebuf[buf][i].w2 & 0x7fff;
    switch(signs) {
    case DRAW_SAVE:
#ifdef SEEREPLAY
      fprintf(stderr,"(%3d) Draw:     %6d %6d\n",nsave,m1,m2);
#endif
      mgxdraw(m1, m2);
      fillend = nsave+1;
      break;
    case RELO_SAVE:
#ifdef SEEREPLAY
      fprintf(stderr,"(%3d) Relocate: %6d %6d\n",nsave,m1,m2);
#endif
      mgxrelocate(m1, m2);
      break;
    case DOT_SAVE:
#ifdef SEEREPLAY
      fprintf(stderr,"(%3d) Dot:      %6d %6d\n",nsave,m1,m2);
#endif
      mgxrelocate(m1, m2);
      mgxdraw(m1, m2);
      break;
    case ETC_SAVE:
      switch(m1) {
      case END_SAVE:
	break;
      case FINI_SAVE:
	break;
      case LW_SAVE:
#ifdef SEEREPLAY
      fprintf(stderr,"(%3d) Lweight:  %6d\n",nsave,m2);
#endif
	mgxlweight(m2);
	break;
      case LT_SAVE:
#ifdef SEEREPLAY
      fprintf(stderr,"(%3d) Ltype:    %6d\n",nsave,m2);
#endif
	mgxltype(m2);
	break;
      case CLR_SAVE:
#ifdef SEEREPLAY
      fprintf(stderr,"(%3d) Color:    %6d\n",nsave,m2);
#endif
	c = m2;            /* Keep around in case of oncoming palette save */
	mgxselcol(m2);
	break;
      case RED_SAVE:
#ifdef SEEREPLAY
      fprintf(stderr,"(%3d) Red:      %6d %3d\n",nsave,m2,c);
#endif
	pal[0] = m2;
	break;
      case GRN_SAVE:
#ifdef SEEREPLAY
      fprintf(stderr,"(%3d) Green:    %6d %3d\n",nsave,m2,c);
#endif
	pal[1] = m2;
	break;
      case BLUE_SAVE:
#ifdef SEEREPLAY
      fprintf(stderr,"(%3d) Blue:     %6d %3d\n",nsave,m2,c);
#endif
	pal[2] = m2;
	mgxpaladd(c, pal);
	break;
      case STR_SAVE:
#ifdef SEEREPLAY
      fprintf(stderr,"(%3d) String:   %6d '%s'\n",nsave,m2,&stringbuf[m2]);
#endif
	mgxstring(xlast, ylast, &stringbuf[m2]);
	break;
      case FILL_SAVE:
#ifdef SEEREPLAY
      fprintf(stderr,"(%3d) Fill:     %6d %3d\n",nsave,m2%1024,m2/1024);
#endif
	mgxsetfill(m2%1024, m2/1024);
	break;
      }
    }
  }
  REPLAY = 0;
  nsave = NSAVE;
  return (0); 
}

mgxsaveit(signs, m1, m2)
int signs, m1, m2;
{
  int buf, i, s1, s2;
  buf = nsave / MAXVEC;

/* First check to if we're beyond the last save buffer pointer */
  if(buf > MAXBUF-1) return(-1);

/* Get more memory if necessary */
  if(savebuf[buf] == NULL) {
    savebuf[buf] = (struct mgxsavebuf *)
      malloc(MAXVEC*sizeof(struct mgxsavebuf));
  }
  i = nsave % MAXVEC;
  s1 = (signs & 1) << 15;
  s2 = (signs / 2) << 15;
  savebuf[buf][i].w1 = (unsigned short int)(s1 + (m1 & 0x7fff));
  savebuf[buf][i].w2 = (unsigned short int)(s2 + (m2 & 0x7fff));
  nsave++;
  return(0);
}

#include <X11/Xutil.h>
#include <X11/keysym.h>

/*
 * Interrupt handler for X Events
 */

char *mgxrefresh(string, len)
char *string;
int len;
{
  XEvent event;
  XExposeEvent *expw  = (XExposeEvent *)&event;
  XButtonEvent *but = (XButtonEvent *)&event;
  XKeyEvent *key = (XKeyEvent *)&event;
  XPointerMovedEvent *pmove = (XPointerMovedEvent *)&event;
  int redrawn;

  Window wmouse;
  KeySym ks;
  int wmask;
#define KEYLEN 10
  char keystring[KEYLEN];
  XComposeStatus compose_status;
  int i;
  char keychar;
  char outbuf[32];

  redrawn = 0;

  if(MGXK) {
/* Pull out only those events for the Mongo window */
    while( XCheckWindowEvent(mgxdpy, mgxwin, mgxeventmask, &event)==True ) {

/* Switch on the type of event */
      switch((int)event.type) {
	
/* 
 * WINDOW EXPOSURE EVENT
 */
      case Expose:

/* If the window was resized, update the subwindow sizes */
	if(redrawn) break;

/*
	fprintf(stderr,"Expose event %d %d %d %d\n",expw->x,expw->y,expw->width,expw->height);
*/
	XGetGeometry(mgxdpy, mgxwin, 
		     &mgxinfowin,&mgxinfox,&mgxinfoy,&mgxinfow,&mgxinfoh,
		     &mgxinfob,&mgxinfod);
	if(mgxinfow != width || mgxinfoh != height) {
	  XClearWindow(mgxdpy,mgxwin);
	  width = mgxinfow;
	  height = mgxinfoh;
	  i = minify == 0 ? 1 : ABS(minify);
	  mgoxwinresize(i*width, i*height);
	}
	mgxreplay();
	redrawn = 1;
	break;
/* 
 * BUTTON PRESSED: A mouse button was pressed
 */
      case ButtonPress:
	mousex = but->x;
	mousey = height - but->y - 1;

/* 
 * BUTTON RELEASED: The button was released
 */
      case ButtonRelease:
/* waiting_for_cursor means save mouse position */
        if(waiting_for_cursor != 0) {
          *outbuf = but->button;
	  waiting_for_cursor = 0;
	}

/*
 * KEY PRESSED: A Keyboard key was pressed
 */
      case KeyPress:
	
	i = XLookupString(key,keystring,KEYLEN,&ks,&compose_status);

/* If a shift (modifier) key, ignore it */
	if(IsModifierKey(ks)) break;

/* If an arrow key, move the mouse */
	if(IsCursorKey(ks)) {
	  XQueryPointer(mgxdpy, mgxwin, &mgxinfowin,&wmouse,
			&mgxinfox,&mgxinfoy,&mousex,&mousey,&wmask);
	  if(ks == XK_Left) mousex--;
	  else if(ks == XK_Right) mousex++;
	  else if(ks == XK_Down) mousey++;
	  else if(ks == XK_Up) mousey--;
	  XWarpPointer(mgxdpy, None, mgxwin, 0,0,0,0,mousex,mousey);
	  if(waiting_for_cursor == 1) {
	    mgxnewcoords(0,mousex, mousey);
	  } else if (waiting_for_cursor == 2) {
	    mgxnewcoords(1,mousex, mousey);
	  }
	  XFlush(mgxdpy);
	  break;
	}

/* If the length is equal to zero, quit */
	if(i == 0) break;

/* Get the character typed */
	keychar = keystring[0];

/* waiting_for_cursor means save mouse position */
	if(waiting_for_cursor != 0) {
	  *outbuf = keychar;
	  mousex = key->x;
	  mousey = height - key->y - 1;
	  waiting_for_cursor = 0;
	}
	break;
/* 
 * MOUSE MOVED: Mouse movement
 */
      case MotionNotify:
	XQueryPointer(mgxdpy,mgxwin,&mgxinfowin,&wmouse,
		      &mgxinfox,&mgxinfoy,&mousex,&mousey,&wmask);
	if(waiting_for_cursor == 1) {	
	  mgxnewcoords(0,mousex, mousey);
	} else if (waiting_for_cursor == 2) {
	  mgxnewcoords(1,mousex, mousey);
	}
	break;
      }
    }
  }
  return(NULL);
}

mgxchangeclrs(dp, map, color, n)
Display *dp;
Colormap map;
XColor *color;
int n;
{
  int i;
  XColor test;
  for(i=0; i<n; i++) mgxsetcolor(dp, vclass, map, color+i);
  /*
  for(i=0; i<n; i++) {
    test.pixel = color[i].pixel;
    XQueryColor(dp, map, &test);
    printf("Color %2d: %5d %7d %7d %7d\n", i, test.pixel, test.red, test.green, test.blue);
  }
  */
}
